import pytest
from .util import determine_html_url


def test_determine_html_url():
    assert determine_html_url([]) is None


@pytest.mark.parametrize(
    "urls,expected",
    [
        (
            [
                {"mediaType": "unknown", "href": "http://wrong.test"},
                {"mediaType": "text/html", "href": "http://right.test"},
            ],
            "http://right.test",
        ),
        ([{"href": "http://right.test"}], "http://right.test"),
    ],
)
def test_determine_html_url_result(urls, expected):
    assert determine_html_url(urls) == expected



"""
This package contains helper methods allowing one to write nicer
code in a combination of cattle_grid, muck_out, and magic injection.

## Examples

We now illustrate the usage of these methods

### Fetch

The annotations starting with fetch are available whenever, you have
a configured broker and activity exchange. Their usage is as

```python
async def my_method(fetcher: FetchObject):
    obj = await fetcher(
        "http://my.example/actor/id",
        "http://other.example/object/id"
    )
```

We note that the actor id must belong to an actor managed by cattle_grid.

### Parsed

The Parsed- annotations are used in processing, i.e.

```python
@subscribe("some_routing_key")
async def process_actor(actor: ParsedActor):
    ...
```

where one is assumed to process messages of type
[ActivityMessage][cattle_grid.model.ActivityMessage], or something
inheriting from it.

### Transform

```python
@extension.transform(inputs=["parsed"], outputs=[...])
async def transform(
    actor: TransformedActor
) -> dict[str, Any]:
    ...
```

"""

from collections.abc import Awaitable, Callable
from typing import Annotated
from fast_depends import Depends

from muck_out.cattle_grid.transform import (
    transform_activity,
    transform_actor,
    transform_collection,
    transform_embedded_object,
    transform_object,
)
from muck_out.types import Activity, Actor, Object, Collection

from .fetch import (
    fetch_activity_builder,
    fetch_actor_builder,
    fetch_collection_builder,
    fetch_object_builder,
)

from .methods import (
    get_activity,
    get_actor,
    get_collection,
    get_embedded_actor,
    get_embedded_object,
    get_object,
)


ParsedActivity = Annotated[Activity | None, Depends(get_activity)]
"""Returns the parsed activity from the muck_out extension.

This dependency works for methods that would normally receive
a [ActivityMessage][cattle_grid.model.ActivityMessage], e.g.
the `incoming.*` and `outgoing.*` subscriber of cattle_grid.
"""

ParsedActor = Annotated[Actor | None, Depends(get_actor)]
"""Returns the parsed actor from the muck_out extension.

This dependency works for methods that would normally receive
a [ActivityMessage][cattle_grid.model.ActivityMessage], e.g.
the `incoming.*` and `outgoing.*` subscriber of cattle_grid.
"""

ParsedCollection = Annotated[Collection | None, Depends(get_collection)]
"""Returns the parsed collection from the muck_out extension.

This dependency works for methods that would normally receive
a [ActivityMessage][cattle_grid.model.ActivityMessage], e.g.
the `incoming.*` and `outgoing.*` subscriber of cattle_grid.
"""

ParsedEmbeddedObject = Annotated[Object | None, Depends(get_embedded_object)]
"""Returns the parsed embedded object from an activity from the muck_out extension.

This dependency works for methods that would normally receive
a [ActivityMessage][cattle_grid.model.ActivityMessage], e.g.
the `incoming.*` and `outgoing.*` subscriber of cattle_grid.
"""

ParsedEmbeddedActor = Annotated[Actor | None, Depends(get_embedded_actor)]
"""Returns the parsed embedded actor from an activity from the muck_out extension."""

ParsedObject = Annotated[Object | None, Depends(get_object)]
"""Returns the parsed object from the muck_out extension.

This dependency works for methods that would normally receive
a [ActivityMessage][cattle_grid.model.ActivityMessage], e.g.
the `incoming.*` and `outgoing.*` subscriber of cattle_grid.
"""

FetchActivity = Annotated[
    Callable[[str, str], Awaitable[Activity | None]], Depends(fetch_activity_builder)
]
"""Returns the activity after fetching it"""
FetchActor = Annotated[
    Callable[[str, str], Awaitable[Actor | None]], Depends(fetch_actor_builder)
]
"""Returns the actor after fetching it"""
FetchCollection = Annotated[
    Callable[[str, str], Awaitable[Collection | None]],
    Depends(fetch_collection_builder),
]
"""Returns the collection after fetching it"""
FetchObject = Annotated[
    Callable[[str, str], Awaitable[Object | None]], Depends(fetch_object_builder)
]
"""Returns the object after fetching it"""

TransformedActivity = Annotated[Activity | None, Depends(transform_activity)]
"""Inside of a running transformer, returns the Activity as transformed by muck_out"""
TransformedActor = Annotated[Actor | None, Depends(transform_actor)]
"""Inside of a running transformer, returns the Actor as transformed by muck_out"""
TransformedCollection = Annotated[Collection | None, Depends(transform_collection)]
"""Inside of a running transformer, returns the Collection as transformed by muck_out"""
TransformedEmbeddedObject = Annotated[Object | None, Depends(transform_embedded_object)]
"""Inside of a running transformer, returns the EmbeddedObject as transformed by muck_out"""
TransformedObject = Annotated[Object | None, Depends(transform_object)]
"""Inside of a running transformer, returns the Object as transformed by muck_out"""



from typing import Any


def determine_html_url(urls: list[dict[str, Any]]) -> str | None:
    for url in urls:
        if url.get("mediaType") == "text/html":
            if url.get("href"):
                return url.get("href")
    for url in urls:
        if url.get("href"):
            return url.get("href")
    return None



from muck_out.process import normalize_object
from muck_out.types import ObjectStub

from muck_out.testing.examples import sample_object


def test_normalize_object():
    result = normalize_object(sample_object)

    assert result.id == sample_object["id"]
    assert result.context == sample_object["context"]


def test_object_stub():
    result = ObjectStub.model_validate(sample_object)

    assert result.id == sample_object["id"]
    assert result.context == sample_object["context"]



"""
This package contains tools to turn ActivityPub messages
into something cleaner. Its takes include

- Normalization
- Validation


"""

import logging

from pydantic import BaseModel, Field

from .types import Activity, Object, Actor, Collection

from .process import (
    normalize_activity,
    normalize_object,
    normalize_collection,
    normalize_actor,
)


logger = logging.getLogger(__name__)


class NormalizationResult(BaseModel):
    object: Object | None = Field(default=None)
    activity: Activity | None = Field(default=None)

    actor: Actor | None = Field(default=None)
    collection: Collection | None = Field(default=None)

    embedded_object: Object | None = Field(default=None)
    embedded_actor: Actor | None = Field(default=None)


def normalize_data(
    data: dict, actor_id: str | None = None, drop_context: bool = False
) -> NormalizationResult:
    """Normalizes the object"""

    def result_or_none(func):
        try:
            result = func(data)
            if drop_context:
                result.field_context = None
            return result
        except Exception as e:
            logger.debug(e)

            return None

    try:
        activity = normalize_activity(data, actor_id=actor_id)
        if drop_context:
            activity.field_context = None
    except Exception as e:
        logger.debug(e)
        activity = None

    if activity and isinstance(activity.object, Object):
        embedded_object = activity.object
        activity.object = embedded_object.id
        if drop_context:
            embedded_object.field_context = None
    else:
        embedded_object = None

    embedded_actor = None

    if activity and activity.type == "Update":
        if activity.actor == activity.object:
            try:
                embedded_actor = normalize_actor(data.get("object"))  # type: ignore
                if drop_context and embedded_actor:
                    embedded_actor.field_context = None

            except Exception:
                ...

    return NormalizationResult(
        object=result_or_none(normalize_object),
        activity=activity,
        embedded_object=embedded_object,
        actor=result_or_none(normalize_actor),
        collection=result_or_none(normalize_collection),
        embedded_actor=embedded_actor,
    )



from typing import Annotated
from pydantic import BaseModel, ConfigDict, Field

from muck_out.pydantic_types import HtmlStringOrNone


class ActorHeaderInfo(BaseModel):
    """Derived information suitable to be displayed in a header line."""

    model_config = ConfigDict(
        serialize_by_alias=True,
        extra="forbid",
        field_title_generator=lambda field_name, field_info: "",
        validate_by_name=True,
    )

    id: str = Field(
        examples=["https://actor.example/alice/some/id"],
        description="id of the actor, can be assumed to be globally unique.",
    )

    avatar_url: Annotated[
        str | None,
        Field(
            default=None,
            examples=["https://actor.example/static/alice-avatar.png"],
            description="The url of the avatar to use for the actor",
            alias="avatarUrl",
        ),
    ]

    name: HtmlStringOrNone = Field(
        default=None,
        examples=["Alice"],
        description="Display name of the actor",
    )

    identifier: str = Field(
        examples=["acct:alice@actor.example"],
        description="Identifier to display for the actor",
    )

    html_url: Annotated[
        str | None,
        Field(
            default=None,
            examples=["https://actor.example/@alice"],
            description="Location of a html representation of the actor",
            alias="htmlUrl",
        ),
    ]



from pathlib import Path
import click


from .types import Actor, Activity, Object, Collection
from .derived import ActorHeaderInfo, ObjectMetaInfo
from .sub_types import Hashtag, Mention, PropertyValue
from .schemas import to_json_schema


@click.group
def main(): ...


@main.command()
@click.option("--path", default="docs/schemas/")
def schemas(path: str):
    Path(path).mkdir(exist_ok=True, parents=True)
    for obj, name in [
        (Actor, "actor"),
        (Activity, "activity"),
        (Object, "object"),
        (Collection, "collection"),
        (ActorHeaderInfo, "actor-header-info"),
        (ObjectMetaInfo, "object-meta-info"),
        (Hashtag, "hashtag"),
        (Mention, "mention"),
        (PropertyValue, "property-value"),
    ]:
        to_json_schema(obj, f"{path}/{name}.json")  # type: ignore


if __name__ == "__main__":
    main()



import re
import pytest

from muck_out.testing import load_asset
from muck_out.types import Actor
from muck_out.sub_types import PropertyValue

from muck_out.testing.examples import mitra_actor, actor_example

from .actor import normalize_actor, actor_stub


def test_normalize_actor_failure():
    assert normalize_actor({}) is None


actor_base = {
    "@context": "https://www.w3.org/ns/activitystreams",
    "type": "Person",
    "id": "https://example.com/users/alice",
}


@pytest.mark.parametrize(
    "data",
    [
        {
            "@context": "https://www.w3.org/ns/activitystreams",
            "type": "Person",
            "id": "https://example.com/users/alice",
            "inbox": "https://example.com/users/alice/inbox",
        },
        {
            "@context": "https://www.w3.org/ns/activitystreams",
            "type": "Person",
            "id": "https://example.com/users/alice",
            "inbox": "https://example.com/users/alice/inbox",
            "icon": {
                "type": "Image",
                "url": "https://example.com/users/alice/icon.png",
            },
        },
    ],
)
def test_normalize_actor(data):
    assert isinstance(normalize_actor(data), Actor)


def to_snake_case(string: str):
    return re.sub("([A-Z]+)", r"_\1", string).lower()


@pytest.mark.parametrize(
    "field, value, mapped",
    [
        ("summary", "<p>Hello, world!</p>", "<p>Hello, world!</p>"),
        (
            "summary",
            "<p>Test<script>alert('hi')</script></p>",
            "<p>Test</p>",
        ),
        ("name", "Alice", "Alice"),
        (
            "alsoKnownAs",
            ["https://example.com/users/alice"],
            ["https://example.com/users/alice"],
        ),
    ],
)
def test_normalize_actor_field(field: str, value: str, mapped: str):
    data = actor_base.copy()
    data[field] = value
    actor = actor_stub(data)
    assert getattr(actor, to_snake_case(field)) == mapped


def test_actor_import():
    result = actor_stub(mitra_actor)

    assert result.url
    assert result.preferred_username == "admin"
    assert result.identifiers == ["acct:admin@mitra", "http://mitra/users/admin"]

    assert result.following == mitra_actor["following"]
    assert result.followers == mitra_actor["followers"]

    assert result.url == [
        {
            "href": "http://mitra/users/admin",
            "type": "Link",
        },
    ]


def test_actor_import_icon():
    result = actor_stub(actor_example)

    assert result.icon


def test_actor_with_property_values():
    property_value = {
        "type": "PropertyValue",
        "name": "Blog",
        "value": "http://blog.test",
    }
    actor = {**actor_example}
    actor["attachment"] = [property_value]

    stub = actor_stub(actor)

    assert stub.attachment
    assert len(stub.attachment) == 1

    ele = stub.attachment[0]

    assert isinstance(ele, PropertyValue)


def test_actor_with_none_attachment():
    actor = {**actor_example}
    actor["attachment"] = None

    stub = actor_stub(actor)

    assert stub


def test_peertube_account():
    actor = load_asset("peertube_account")

    normalize_actor(actor)



from pydantic import BaseModel

from . import ActorAttachment, PropertyValue


class Helper(BaseModel):
    value: ActorAttachment


def test_unknown():
    result = Helper.model_validate({"value": {"type": "Other"}})

    assert isinstance(result.value, dict)


def test_property_value():
    property_value = {"type": "PropertyValue", "name": "key", "value": "value"}
    result = Helper.model_validate({"value": property_value})

    assert isinstance(result.value, PropertyValue)


def test_property_value_as_value():
    property_value = PropertyValue.model_validate(
        {"type": "PropertyValue", "name": "key", "value": "value"}
    )
    result = Helper.model_validate({"value": property_value})

    assert isinstance(result.value, PropertyValue)


def test_property_value_removes_html():
    property_value = {
        "type": "PropertyValue",
        "name": "Blog",
        "value": '<a href="http://value.example/" target="_blank" rel="nofollow noopener noreferrer me" translate="no"><span class="invisible">http://</span><span class="">value.example/</span><span class="invisible"></span></a>',
    }
    result = Helper.model_validate({"value": property_value})
    assert isinstance(result.value, PropertyValue)
    assert result.value.value == "http://value.example/"



from pydantic import BaseModel
import pytest

from .tag import Hashtag, Mention, TagType


class Helper(BaseModel):
    value: TagType


@pytest.mark.parametrize(
    "value,expected_type",
    [
        ({"type": "Unknown"}, dict),
        ({"type": "Hashtag", "name": "#cow"}, Hashtag),
        (Hashtag(name="#cow"), Hashtag),
        ({"type": "Mention", "href": "http://remote.test/actor"}, Mention),
        (Mention(href="http://remote.test/actor"), Mention),
    ],
)
def test_tag_type(value, expected_type):
    result = Helper.model_validate({"value": value})

    assert isinstance(result.value, expected_type)



from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, BeforeValidator
from typing import Annotated

from muck_out.pydantic_types import IdOrNone, TransformToListOfUris
from muck_out.transform.list_utils import list_from_value


class CommonAll(BaseModel):
    """Common base for all ActivityPub objects.

    Note `@context` is normalized to be a list, i.e

    ```pycon
    >>> CommonAll.model_validate({"@context":
    ...     "https://www.w3.org/ns/activitystreams"})
    CommonAll(field_context=['https://www.w3.org/ns/activitystreams'], id=None, type=None)

    ```
    """

    model_config = ConfigDict(serialize_by_alias=True)

    field_context: Annotated[
        list[str | dict] | None, BeforeValidator(list_from_value)
    ] = Field(
        default=["https://www.w3.org/ns/activitystreams"],
        alias="@context",
        examples=[
            ["https://www.w3.org/ns/activitystreams"],
            ["https://www.w3.org/ns/activitystreams", {"Hashtag": "as:Hashtag"}],
        ],
        description="The Json-LD context",
    )
    id: str | None = Field(
        default=None,
        examples=["https://actor.example/some_id"],
        description="id of the activity or object, can be assumed to be globally unique. Some activities such as a Follow request will require an id to be valid. Servers may assume an id to be required. As assigning an id is 'trivial', one should assign one.",
    )
    type: IdOrNone = Field(
        default=None,
        examples=[
            "Follow",
            "Accept",
            "Create",
            "Undo",
            "Like",
            "Note",
            "Actor",
            "Collection",
        ],
        description="Type of the activity or activity. Side effects of this activity are determine by this type.",
    )


class Common(CommonAll):
    to: TransformToListOfUris = Field(
        default=[],
        examples=[
            ["https://bob.example"],
            ["https://alice.example", "https://bob.example"],
        ],
        description="Array of actors this activity or object is addressed to. It is sane to assume that an activity is addressed to at least one person.",
    )
    cc: TransformToListOfUris = Field(
        default=[],
        examples=[
            ["https://bob.example"],
            ["https://alice.example", "https://bob.example"],
        ],
        description="Array of actors this activity or object is carbon copied to.",
    )
    published: datetime | None = Field(
        default=None,
        description="Moment of this activity or object being published",
    )



import json
from typing import Any


def load_asset(name: str) -> dict[str, Any]:
    """
    Loads a test asset.

    ```python
    >>> result = load_asset("mastodon_actor_update")
    >>> result["type"]
    'Update'

    ```
    """
    with open(f"docs/assets/{name}.json") as fp:
        return json.load(fp)



from cattle_grid.extensions import Extension
from cattle_grid.dependencies.transformer import ActorTransforming

from muck_out.extension.config import MuckOutConfiguration

from .. import normalize_data

extension = Extension(
    name="muck out", module=__name__, config_class=MuckOutConfiguration
)


@extension.transform(inputs=["raw"], outputs=["parsed"])
async def muck_out(data: dict, actor: ActorTransforming, config: extension.Config):  # type: ignore
    actor_id = actor.actor_id if actor else None

    normalized = normalize_data(
        data["raw"], actor_id=actor_id, drop_context=config.drop_context
    )

    return {"parsed": normalized.model_dump(by_alias=True, exclude_none=True)}



from bovine.activitystreams.utils import id_for_object
from muck_out.transform.utils import sanitize_html


def safe_html(x) -> str | None:
    if x is None:
        return None
    if isinstance(x, list):
        if len(x) == 0:
            return None
        x = x[0]
    return sanitize_html(x)


def safe_id_for_object(x):
    if isinstance(x, list):
        if len(x) == 0:
            return None
        x = x[0]
    return id_for_object(x)


def single_element_list_to_element(x: str | None | list[str]) -> str | None:
    """Returns a sanitized version of the result:

    ```python
    >>> single_element_list_to_element(None)

    >>> single_element_list_to_element("Note")
    'Note'

    >>> single_element_list_to_element([])

    >>> single_element_list_to_element(["Note"])
    'Note'

    >>> single_element_list_to_element(["Note", "Article"])
    Traceback (most recent call last):
    ...
    ValueError: Cannot handle lists of more than 1 element

    """

    if isinstance(x, list):
        if len(x) > 1:
            raise ValueError("Cannot handle lists of more than 1 element")
        if len(x) == 1:
            return x[0]
        return None
    return x



from typing import Annotated

from pydantic import AfterValidator


def at_least_one_element(value: list) -> list:
    if len(value) == 0:
        raise ValueError("List must contain at least one value")
    return value


ListOfStringsWithAtLeastOneElement = Annotated[
    list[str], AfterValidator(at_least_one_element)
]



from .attachment import PropertyValue, ActorAttachment
from .tag import Mention, Hashtag, ObjectLink, TagType

__all__ = [
    "PropertyValue",
    "ActorAttachment",
    "Mention",
    "Hashtag",
    "ObjectLink",
    "TagType",
]



import pytest

from .utils import sanitize_html


@pytest.mark.parametrize(
    "unchanged",
    [
        '<a href="http://actor.test/" class="mention" rel="noopener noreferrer">@actor</a>',
        '<a href="http://actor.test/" class="hashtag" rel="noopener noreferrer">@actor</a>',
        '<a href="http://actor.test/" class="mention hashtag" rel="noopener noreferrer">@actor</a>',
    ],
)
def test_sanitize_html_class_kept(unchanged):
    assert sanitize_html(unchanged) == unchanged


def test_santize_html_class_removed():
    bad = '<a href="http://actor.test/" class="weird" rel="noopener noreferrer">@actor</a>'

    assert (
        sanitize_html(bad)
        == '<a href="http://actor.test/" rel="noopener noreferrer">@actor</a>'
    )



from pydantic import Field, BeforeValidator
from typing import Any, Annotated

from muck_out.types.common import CommonObject
from muck_out.pydantic_types import HtmlStringOrNone, IdFieldOrNone, UrlList
from muck_out.transform.attachment import transform_attachments

from .common import Common


class ObjectStub(Common, CommonObject):
    """Stub object"""

    attributed_to: IdFieldOrNone = Field(
        default=None,
        examples=["https://actor.example/"],
        description="id of the actor that authored this object",
        alias="attributedTo",
    )
    content: HtmlStringOrNone = Field(
        default=None, description="The content of the object"
    )

    attachment: Annotated[
        list[dict[str, Any]], BeforeValidator(transform_attachments)
    ] = Field(
        default=[],
        description="A list of objects that are attached to the original object",
    )

    url: UrlList = Field(
        default=[],
        description="A list of urls that expand on the content of the object",
    )
    sensitive: bool | None = Field(
        None,
        description="Marks the object as sensitive. Currently, used by everyone, a better way would be an element of the tag list that labels the object as sensitive due a reason",
    )



import pytest
from muck_out.testing import load_asset
from .activity import activity_stub, normalize_activity


def test_mastodon_like():
    mastodon_like = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "https://mastodon.social/users/the_milkman#likes/251507741",
        "type": "Like",
        "actor": "https://mastodon.social/users/the_milkman",
        "object": "https://dev.bovine.social/html_display/object/01999580-e682-799a-8a43-ae9f5742d148",
    }

    actor_id = "http://local.test/actor/id"

    result = activity_stub(mastodon_like, actor_id=actor_id)

    assert result.to == [actor_id]


@pytest.mark.parametrize(
    "name", ["fep_c0e0_emoji_react", "fep_c0e0_custom_emoji_react"]
)
def test_emoji_reaction(name):
    """Example from [FEP-c0e0: Emoji reactions](https://fediverse.codeberg.page/fep/fep/c0e0/)"""
    emoji_react = load_asset(name)

    result = activity_stub(emoji_react)

    assert result.type == "Like"


def test_internal_cattle_grid_test():
    activity = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "type": "Accept",
        "actor": "http://cattle_grid/actor/Twle_a-3qNL7cHZra5c8Kg",
        "to": ["http://cattle_grid/actor/484KcCWjPNCVl0BQFenJiQ"],
        "published": "2025-11-16T14:56:13Z",
        "object": "follow:cf8684cc-5738-4862-b84f-80e44721572b",
        "id": "http://cattle_grid/object/019a8d2a-a69c-7a9a-addd-82f0d34d5a6a",
    }

    result = normalize_activity(activity)

    assert result
    assert result.type == "Accept"



from datetime import datetime
from typing import Annotated
from pydantic import BaseModel, BeforeValidator, ConfigDict, Field

from muck_out.pydantic_types import HtmlStringOrNone, IdFieldOrNone
from muck_out.sub_types import TagType
from muck_out.transform.list_utils import transform_to_list


class CommonObject(BaseModel):
    model_config = ConfigDict(
        serialize_by_alias=True,
        validate_by_name=True,
        field_title_generator=lambda field_name, field_info: "",
    )

    updated: datetime | None = Field(
        default=None,
        description="Moment of this object being updated",
    )
    summary: HtmlStringOrNone = Field(
        default=None,
        description="The summary of the object",
    )
    name: HtmlStringOrNone = Field(
        default=None,
        description="The name of the object",
    )
    in_reply_to: IdFieldOrNone = Field(
        None,
        description="The object being replied to. Currently a string. Not sure if this is what I want.",
        alias="inReplyTo",
    )
    context: IdFieldOrNone = Field(None, description="The context of the object")
    tag: Annotated[list[TagType], BeforeValidator(transform_to_list)] = Field(
        default=[],
        description="A list of objects that expand on the content of the object",
    )



from .activity import normalize_activity, activity_stub
from .actor import actor_stub, normalize_actor
from .collection import normalize_collection, collection_stub
from .object import object_stub, normalize_object

__all__ = [
    "activity_stub",
    "actor_stub",
    "collection_stub",
    "object_stub",
    "normalize_activity",
    "normalize_object",
    "normalize_actor",
    "normalize_collection",
]



from .actor import ActorStub


def test_identifiers():
    stub = ActorStub.model_validate({"identifiers": "acct:some@domain"})

    assert stub.identifiers == ["acct:some@domain"]


def test_manually_approves_followers():
    stub = ActorStub.model_validate({"manuallyApprovesFollowers": True})

    assert stub



from typing import Annotated, Literal
from pydantic import BaseModel, BeforeValidator, Discriminator, Field, Tag


class Mention(BaseModel):
    """Represents a mention

    ```python
    >>> m = Mention(href="http://actor.example/alice", name="@alice@actor.example")
    >>> m.model_dump()
    {'type': 'Mention',
        'href': 'http://actor.example/alice',
        'name': '@alice@actor.example'}

    ```
    """

    type: Literal["Mention"] = Field(default="Mention")
    href: str = Field(
        description="The location the mentioned party can be retrieved at. In the Fediverse usually an actor URI"
    )
    name: str | None = Field(default=None)


class Hashtag(BaseModel):
    """Represents a hashtag

    ```python
    >>> m = Hashtag(name="#cow")
    >>> m.model_dump(exclude_none=True)
    {'type': 'Hashtag', 'name': '#cow'}

    ```
    """

    type: Literal["Hashtag"] = Field(default="Hashtag")
    href: str | None = Field(
        default=None, description="A location related to the hashtag"
    )
    name: str = Field(description="The actual hashtag", examples=["#cow"])


class ObjectLink(BaseModel):
    """Represents a [FEP-e232: Object Link](https://fediverse.codeberg.page/fep/fep/e232/)"""

    type: Literal["Link"] = Field(default="Link")
    href: str = Field(
        examples=["http://remote.example/object/12345"],
        description="The location of the object",
    )
    mediaType: str = Field(
        examples=[
            '''application/ld+json; profile="https://www.w3.org/ns/activitystreams"''',
            "application/activity+json",
        ],
        description="The media type of the object",
    )
    name: str | None = Field(
        default=None,
        examples=["RE http://remote.example/object/12345"],
        description="The microsyntax used to represent the object",
    )
    rel: str | None = Field(default=None, description="Relation to the object")


def discriminator_tag(v):
    match v:
        case {"type": "Hashtag"} | Hashtag():
            return "Hashtag"
        case {"type": "Mention"} | Mention():
            return "Mention"
        case ObjectLink() | {"type": "Link", "mediaType": str()}:
            return "ObjectLink"
        case dict():
            return "unknown"
        case str():
            return "string"

    raise Exception


def to_link(href: str) -> dict:
    return {"href": href}


TagType = Annotated[
    Annotated[Hashtag, Tag("Hashtag")]
    | Annotated[Mention, Tag("Mention")]
    | Annotated[ObjectLink, Tag("ObjectLink")]
    | Annotated[dict, Tag("unknown")]
    | Annotated[Annotated[dict, BeforeValidator(to_link)], Tag("string")],
    Discriminator(discriminator_tag),
]



from .collection import CollectionStub


def test_collection_stub_embedded_first():
    first_id = "http://mastodon.example/users/alice/statuses/2343242/replies?first=true"
    data = {
        "id": "http://mastodon.example/users/alice/statuses/2343242/replies",
        "type": "Collection",
        "first": {
            "id": first_id,
            "type": "CollectionPage",
            "next": "http://mastodon.example/users/alice/statuses/2343242/replies?only_other_accounts=true&page=true",
            "partOf": "http://mastodon.example/users/alice/statuses/2343242/replies",
            "items": [],
        },
    }
    result = CollectionStub.model_validate(data)

    assert result.first == first_id


def test_collection_mastodon_outbox_page_type():
    data = {
        "items": ["item1", {"aaa": "bbb"}, {"id": "item2"}],
    }
    result = CollectionStub.model_validate(data)

    assert result.items == ["item1", "item2"]



import pytest
from muck_out.process.collection import collection_stub


def test_collection_mastodon_outbox_type():
    data = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "http://mastodon.example/users/alice/outbox",
        "type": "OrderedCollection",
        "totalItems": 12360,
        "first": "http://mastodon.example/users/alice/outbox?page=true",
        "last": "http://mastodon.example/users/alice/outbox?min_id=0&page=true",
    }

    result = collection_stub(data)

    assert result.id == data["id"]
    assert result.total_items == data["totalItems"]
    assert result.first == data["first"]
    assert result.last == data["last"]


@pytest.mark.parametrize(
    "data",
    [
        {
            "orderedItems": ["item1", {"aaa": "bbb"}, {"id": "item2"}],
        },
        {
            "items": ["item1", {"aaa": "bbb"}, {"id": "item2"}],
        },
    ],
)
def test_mastodon_outbox_page_type(data):
    result = collection_stub(data)

    assert result.items == ["item1", "item2"]


def test_collection_mastodon_replies_type():
    data = {
        "id": "http://mastodon.example/users/alice/statuses/2343242/replies",
        "type": "Collection",
        "first": {
            "type": "CollectionPage",
            "next": "http://mastodon.example/users/alice/statuses/2343242/replies?only_other_accounts=true&page=true",
            "partOf": "http://mastodon.example/users/alice/statuses/2343242/replies",
            "items": ["http://first.test/item"],
        },
    }
    result = collection_stub(data)

    assert result.items == ["http://first.test/item"]
    assert (
        result.next
        == "http://mastodon.example/users/alice/statuses/2343242/replies?only_other_accounts=true&page=true"
    )



import pytest

from muck_out.sub_types import ObjectLink
from .object import object_stub


e232_example1 = {
    "@context": "https://www.w3.org/ns/activitystreams",
    "type": "Note",
    "content": "The bug was reported in #1374",
    "tag": [
        {
            "type": "Link",
            "mediaType": 'application/ld+json; profile="https://www.w3.org/ns/activitystreams"',
            "href": "https://forge.example/tickets/1374",
            "name": "#1374",
        }
    ],
}

e232_example2 = {
    "@context": "https://www.w3.org/ns/activitystreams",
    "type": "Note",
    "content": "This is a quote:<br>RE: https://server.example/objects/123",
    "tag": [
        {
            "type": "Link",
            "mediaType": 'application/ld+json; profile="https://www.w3.org/ns/activitystreams"',
            "href": "https://server.example/objects/123",
            "name": "RE: https://server.example/objects/123",
        }
    ],
}


@pytest.mark.parametrize("example", [e232_example1, e232_example2])
def test_normalize_object(example):
    result = object_stub(example)

    assert len(result.tag) == 1
    tag = result.tag[0]

    assert isinstance(tag, ObjectLink)



import json
from pydantic import BaseModel
from .clean import clean_schema


def to_json_schema(model: BaseModel, filename: str):
    schema = model.model_json_schema()
    schema = clean_schema(schema)
    with open(filename, "w") as f:
        json.dump(schema, f, indent=2)



def clean_definition(schema: dict) -> dict:
    if schema.get("default") is not None:
        return schema

    any_of = schema.get("anyOf", [])

    if len(any_of) != 2:
        return schema

    if any_of[0].get("type") == "null":
        any_other = any_of[1]
    elif any_of[1].get("type") == "null":
        any_other = any_of[0]
    else:
        return schema

    del schema["anyOf"]
    del schema["default"]

    schema = {**schema, **any_other}

    return schema


def clean_schema(schema: dict) -> dict:
    new_schema = {}

    for key, value in schema.items():
        if isinstance(value, dict):
            new_schema[key] = clean_schema(value)
        else:
            if key == "title" and value == "":
                ...
            else:
                new_schema[key] = value

    new_schema = clean_definition(new_schema)

    return new_schema



from pydantic import Field

from muck_out.types.common import CommonActor
from muck_out.pydantic_types.list_utils import ListOfStringsWithAtLeastOneElement

from .common import CommonAll


class Actor(CommonAll, CommonActor):
    """Describes an ActivityPub actor"""

    type: str = Field(
        examples=["Person", "Service", "Application"],
        description="""The type of Actor""",
    )

    icon: dict | None = Field(
        None,
        examples=[{"type": "Image", "url": "https://actor.example/icon.png"}],
        description="The avatar of the actor",
    )

    also_known_as: list[str] | None = Field(
        None,
        examples=[["https://alice.example", "https://alice.example/profile"]],
        alias="alsoKnownAs",
        description="Other uris associated with the actor",
    )

    preferred_username: str | None = Field(
        None, examples=["john"], alias="preferredUsername"
    )

    identifiers: ListOfStringsWithAtLeastOneElement = Field(
        default=[], description="An ordered list of identifiers"
    )



import logging
from typing import Any
from urllib.parse import urlparse

from muck_out.types import ActorStub, Actor

logger = logging.getLogger(__name__)


def calculate_identifiers(stub: ActorStub) -> list[str]:
    if not stub.id:
        return []
    try:
        domain = urlparse(stub.id).netloc

        if stub.preferred_username:
            return [f"acct:{stub.preferred_username}@{domain}", stub.id]

        return [stub.id]
    except Exception:
        return []


def actor_stub(data: dict[str, Any]) -> ActorStub:
    """Returns the stub actor"""

    icon = data.get("icon")
    if isinstance(icon, list):
        if len(icon) > 0:
            data["icon"] = icon[0]
        else:
            data["icon"] = None

    stub = ActorStub.model_validate(data)

    if data.get("identifiers") is None:
        stub.identifiers = calculate_identifiers(stub)

    return stub


def normalize_actor(data: dict[str, Any]) -> Actor | None:
    """Normalizes an ActivityPub actor"""
    stub = actor_stub(data)

    if stub.inbox is None:
        return None

    if stub.identifiers is None or len(stub.identifiers) == 0:
        if stub.id is None:
            return
        stub.identifiers = [stub.id]

    return Actor.model_validate(stub.model_dump(by_alias=True))



from pydantic import Field

from muck_out.pydantic_types import HtmlStringOrNone, IdFieldOrNone

from .common import Common


class ActivityStub(Common):
    """
    This represents a first draft of a json-schema that every activities exchanged between servers MUST satisfy and be able to parse. Here 'being able to parse' means making it to the point, where depending on the type, you decide what side effects to perform.

    Generally, the fields actor, to, and cc (and maybe bcc --- not transported) represent how the message is being delivered. The fields actor, type, object, target, content represent how the message is processed by the server.
    """

    actor: IdFieldOrNone = Field(
        default=None,
        examples=["https://actor.example/"],
        description="id of the actor performing this activity. One can assume that the activity is signed by this actor (in some form).",
    )
    object: IdFieldOrNone = Field(
        default=None, description="The object of the activity"
    )
    target: IdFieldOrNone = Field(
        default=None,
        examples=[
            "https://other.example/target_id",
        ],
        description="The target, not sure if needed, included for completeness",
    )
    content: HtmlStringOrNone = Field(
        default=None,
        examples=["🐮", "❤️"],
        description="The content used for example to represent the Emote for a like",
    )



import pytest

from muck_out.types import ObjectStub, Object
from muck_out.testing import load_asset
from . import object_stub, normalize_object


@pytest.mark.parametrize(
    "data", [{"content": "text"}, {"content": ["text"]}, {"contentMap": {"en": "text"}}]
)
def test_processes_content(data):
    result = object_stub(data)
    assert result.content == "text"


def test_object():
    obj = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "https://comments.bovine.social/pages/aHR0cHM6Ly9ib3ZpbmUuY29kZWJlcmcucGFnZS9jb21tZW50cy8=",
        "type": "Page",
        "to": ["https://www.w3.org/ns/activitystreams#Public"],
        "cc": [],
        "attributedTo": "https://comments.bovine.social/actor/rF4xnx1QraAIU3Krg-2Qzg",
        "name": "Comment Tracking System",
        "summary": None,
        "url": [
            {
                "type": "Link",
                "mediaType": "text/html",
                "href": "https://bovine.codeberg.page/comments/",
            }
        ],
        "context": "https://comments.bovine.social/pages/aHR0cHM6Ly9ib3ZpbmUuY29kZWJlcmcucGFnZS9jb21tZW50cy8=/context",
        "replies": "https://comments.bovine.social/pages/aHR0cHM6Ly9ib3ZpbmUuY29kZWJlcmcucGFnZS9jb21tZW50cy8=/replies",
        "likes": "https://comments.bovine.social/pages/aHR0cHM6Ly9ib3ZpbmUuY29kZWJlcmcucGFnZS9jb21tZW50cy8=/likes",
        "shares": "https://comments.bovine.social/pages/aHR0cHM6Ly9ib3ZpbmUuY29kZWJlcmcucGFnZS9jb21tZW50cy8=/shares",
    }

    stub = object_stub(obj)

    assert isinstance(stub, ObjectStub)

    normalized = normalize_object(obj)

    assert isinstance(normalized, Object)


def test_peertube():
    obj = load_asset("peertube_video")
    old_content = obj.get("content")

    normalize_object(obj)

    assert obj["content"] == old_content


def test_tag_transformation():
    object_stub({"tag": {"type": "Hashtag", "name": "#cow"}})


def test_content_plain():
    obj = {
        "id": "https://domain.test/users/34783",
        "type": "Note",
        "published": "2025-11-11T00:00:00Z",
        "url": "https://domain.test/actor/123",
        "attributedTo": "https://domain.test/actor",
        "to": ["https://www.w3.org/ns/activitystreams#Public"],
        "content": "<p><b>bold</b></p>",
    }

    result = normalize_object(obj)

    assert result.content == obj.get("content")
    assert result.content_plain == "bold"


def test_atlas_object():
    obj = load_asset("atlas_note")

    result = normalize_object(obj)

    assert result.type == "Note"



import pytest

from .common import Common


@pytest.mark.parametrize(
    "to,expected",
    [
        ("http://remote.test/actor", ["http://remote.test/actor"]),
        ({"id": "http://remote.test/actor"}, ["http://remote.test/actor"]),
    ],
)
def test_common_to(to, expected):
    result = Common(to=to)  # type: ignore

    assert result.to == expected



from pydantic import BeforeValidator
from typing import Annotated, Any

from muck_out.transform.utils import remove_html
from muck_out.transform import transform_to_list_of_uris, transform_url

from .methods import safe_html, safe_id_for_object, single_element_list_to_element


HtmlStringOrNone = Annotated[str | None, BeforeValidator(safe_html)]
"""Used for strings, which may contain html"""

IdFieldOrNone = Annotated[str | None, BeforeValidator(safe_id_for_object)]
"""Used for fields meant to contain an id"""


UrlList = Annotated[list[dict[str, Any]], BeforeValidator(transform_url)]
"""Transforms a list of urls"""

TransformToListOfUris = Annotated[list[str], BeforeValidator(transform_to_list_of_uris)]
"""Transforms a list of recipients, ensuring it is a list of URIs"""

PlainText = Annotated[str, BeforeValidator(remove_html)]
"""Ensures a field is plain text, i.e. not containing any html"""

IdOrNone = Annotated[str | None, BeforeValidator(single_element_list_to_element)]
"""Type that ensures a single element list is rendered as an element"""



from .clean import clean_schema


def test_clean_json_schema_one_element():
    schema = {
        "anyOf": [{"type": "string"}, {"type": "null"}],
        "default": None,
        "description": "A location related to the hashtag",
        "title": "Href",
    }

    result = clean_schema(schema)

    assert result == {
        "type": "string",
        "description": "A location related to the hashtag",
        "title": "Href",
    }


def test_clean_json_schema_properties():
    schema = {
        "type": {
            "const": "Hashtag",
            "default": "Hashtag",
            "title": "Type",
            "type": "string",
        },
        "href": {
            "anyOf": [{"type": "string"}, {"type": "null"}],
            "default": None,
            "description": "A location related to the hashtag",
            "title": "Href",
        },
        "name": {
            "description": "The actual hashtag",
            "examples": ["#cow"],
            "title": "Name",
            "type": "string",
        },
    }

    result = clean_schema(schema)

    assert result == {
        "href": {
            "description": "A location related to the hashtag",
            "title": "Href",
            "type": "string",
        },
        "name": {
            "description": "The actual hashtag",
            "examples": ["#cow"],
            "title": "Name",
            "type": "string",
        },
        "type": {
            "const": "Hashtag",
            "default": "Hashtag",
            "title": "Type",
            "type": "string",
        },
    }


def test_clean_attachment():
    schema = {
        "anyOf": [
            {
                "items": {"additionalProperties": True, "type": "object"},
                "type": "array",
            },
            {"type": "null"},
        ],
        "default": None,
        "description": "A list of objects that are attached to the original object",
        "title": "",
    }

    result = clean_schema(schema)

    assert result == {
        "description": "A list of objects that are attached to the original object",
        "items": {"additionalProperties": True, "type": "object"},
        "type": "array",
    }



from pydantic import BaseModel, Field


class MuckOutConfiguration(BaseModel):
    """Configures muck_out"""

    drop_context: bool = Field(
        default=True, description="If set to true sets the `@context` property to None"
    )



from .list_utils import transform_to_list


def transform_attachment(attachment: dict | str) -> dict:
    """Normalizes an attachment"""

    if isinstance(attachment, str):
        return {"href": attachment}

    if attachment.get("type") != "Document":
        return attachment

    media_type = attachment.get("mediaType")
    if media_type is None:
        return attachment

    if media_type.startswith("image/"):
        attachment["type"] = "Image"
    if media_type.startswith("audio/"):
        attachment["type"] = "Audio"
    if media_type.startswith("video/"):
        attachment["type"] = "Video"

    return attachment


def transform_attachments(attachments) -> list[dict] | None:
    list_of_attachments = transform_to_list(attachments)

    return [transform_attachment(attachment) for attachment in list_of_attachments]



"""These are the basic types of objects muck_out transforms things to"""

from .stubs import ActivityStub, ActorStub, ObjectStub, CollectionStub
from .validated import Actor, Activity, Object, Collection

__all__ = [
    "Actor",
    "Activity",
    "Object",
    "Collection",
    "ActorStub",
    "ActivityStub",
    "ObjectStub",
    "CollectionStub",
]



from . import normalize_data
from .testing import load_asset


def test_actor_update():
    actor_update = load_asset("mastodon_actor_update")

    result = normalize_data(actor_update)

    assert result.activity

    assert result.embedded_actor



from typing import Annotated
from pydantic import BaseModel, Discriminator, Field, Tag

from muck_out.pydantic_types import PlainText


class PropertyValue(BaseModel):
    """
    Key value pairs in the attachment of an actor
    as used by Mastodon
    """

    type: str = Field("PropertyValue", description="""Fixed type for serialization""")

    name: PlainText = Field(
        examples=["Pronouns"],
        description="Key of the value",
    )

    value: PlainText = Field(
        examples=["They/them"],
        description="Value",
    )


def discriminator_actor_attachment(v):
    match v:
        case PropertyValue() | {"type": "PropertyValue"}:
            return "PropertyValue"

    return "unknown"


ActorAttachment = Annotated[
    (Annotated[PropertyValue, Tag("PropertyValue")] | Annotated[dict, Tag("unknown")]),
    Discriminator(discriminator_actor_attachment),
]
"""Discriminates between the possible values of an attachment to an actor"""



from typing import Annotated
from pydantic import BaseModel, BeforeValidator, ConfigDict, Field

from muck_out.pydantic_types import HtmlStringOrNone, IdFieldOrNone, UrlList
from muck_out.sub_types import ActorAttachment, TagType
from muck_out.transform.list_utils import transform_to_list


class CommonActor(BaseModel):
    model_config = ConfigDict(
        serialize_by_alias=True,
        validate_by_name=True,
        field_title_generator=lambda field_name, field_info: "",
    )
    inbox: IdFieldOrNone = Field(
        default=None,
        examples=["https://actor.example/inbox"],
        description="The inbox of the actor",
    )

    outbox: IdFieldOrNone = Field(
        default=None,
        examples=["https://actor.example/outbox"],
        description="The outbox of the actor",
    )

    followers: IdFieldOrNone = Field(
        default=None,
        examples=["https://actor.example/followers"],
        description="The followers collection of the actor",
    )

    following: IdFieldOrNone = Field(
        default=None,
        examples=["https://actor.example/following"],
        description="The following collection of the actor",
    )

    summary: HtmlStringOrNone = Field(
        default=None,
        examples=["My Fediverse account"],
        description="Description of the actor",
    )

    name: HtmlStringOrNone = Field(
        default=None,
        examples=["Alice"],
        description="Display name of the actor",
    )

    url: UrlList = Field(
        default=[],
        description="A list of urls connected to the actor",
    )

    attachment: Annotated[list[ActorAttachment], BeforeValidator(transform_to_list)] = (
        Field(
            default=[],
            description="""attachments ... currently used for property values""",
        )
    )

    tag: Annotated[list[TagType], BeforeValidator(transform_to_list)] = Field(
        default=[],
        description="A list of objects that expand on the summary of the actor",
    )

    manually_approves_followers: bool | None = Field(
        default=None,
        alias="manuallyApprovesFollowers",
        description="If set to false, one can assume that te actor automatically replies to follow requests",
    )
